/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*
 * Change History :
 *
*/

#ifndef _H_JFS_XTREE
#define _H_JFS_XTREE

/*
 *      jfs_xtree.h: extent allocation descriptor B+-tree manager
 */

#ifndef _JFS_UTILITY
#include <linux/jfs/jfs_btree.h>
#else
#include "jfs_btree.h"
#endif


/*
 *      extent allocation descriptor (xad)
 */
typedef struct xad {
	unsigned flag:8;	/* 1: flag */
	unsigned rsvrd:16;	/* 2: reserved */
	unsigned off1:8;	/* 1: offset in unit of fsblksize */
	u32 off2;		/* 4: offset in unit of fsblksize */
	unsigned len:24;	/* 3: length in unit of fsblksize */
	unsigned addr1:8;	/* 1: address in unit of fsblksize */
	u32 addr2;		/* 4: address in unit of fsblksize */
} xad_t;			/* (16) */

#define MAXXLEN         ((1 << 24) - 1)

#define XTSLOTSIZE      16
#define L2XTSLOTSIZE    4

/* xad_t field construction */
#define XADoffset(xad, offset64)\
{\
        (xad)->off1 = ((u64)offset64) >> 32;\
        (xad)->off2 = __cpu_to_le32((offset64) & 0xffffffff);\
}
#define XADaddress(xad, address64)\
{\
        (xad)->addr1 = ((u64)address64) >> 32;\
        (xad)->addr2 = __cpu_to_le32((address64) & 0xffffffff);\
}
#define XADlength(xad, length32)        (xad)->len = __cpu_to_le24(length32)

/* xad_t field extraction */
#define offsetXAD(xad)\
        ( ((s64)((xad)->off1)) << 32 | __le32_to_cpu((xad)->off2))
#define addressXAD(xad)\
        ( ((s64)((xad)->addr1)) << 32 | __le32_to_cpu((xad)->addr2))
#define lengthXAD(xad)  __le24_to_cpu((xad)->len)

/* xad list */
typedef struct {
	s16 maxnxad;
	s16 nxad;
	xad_t *xad;
} xadlist_t;

/* xad_t flags */
#define XAD_NEW         0x01	/* new */
#define XAD_EXTENDED    0x02	/* extended */
#define XAD_COMPRESSED  0x04	/* compressed with recorded length */
#define XAD_NOTRECORDED 0x08	/* allocated but not recorded */
#define XAD_COW         0x10	/* copy-on-write */


/* possible values for maxentry */
#define XTROOTINITSLOT_DIR  6
#define XTROOTINITSLOT  10
#define XTROOTMAXSLOT   18
#define XTPAGEMAXSLOT   256
#define XTENTRYSTART    2

/*
 *      xtree page:
 */
typedef union {
	struct xtheader {
		s64 next;	/* 8: */
		s64 prev;	/* 8: */

		u8 flag;	/* 1: */
		u8 rsrvd1;	/* 1: */
		s16 nextindex;	/* 2: next index = number of entries */
		s16 maxentry;	/* 2: max number of entries */
		s16 rsrvd2;	/* 2: */

		pxd_t self;	/* 8: self */
	} header;		/* (32) */

	xad_t xad[XTROOTMAXSLOT];	/* 16 * maxentry: xad array */
} xtpage_t;

#ifndef _JFS_UTILITY
/*
 *      external declaration
 */
extern int xtLookup(
		    struct inode *ip,
		    s64 lstart,
		    int llen,
		    int * pflag, s64 * paddr, int * plen, int flag);

extern int xtLookupPageList(
			    struct inode *ip,
			    s64 lstart,
			    int llen, pxdlist_t * pxdlist, int flag);

extern int xtLookupList(
			struct inode *ip,
			lxdlist_t * lxdlist,
			xadlist_t * xadlist, int flag);

void xtInitRoot( int tid, struct inode *ip);

extern int xtInsert(
		    int tid,
		    struct inode *ip,
		    int xflag, s64 xoff, int xlen, s64 * xaddrp, int flag);

extern int xtExtend(
		    int tid,
		    struct inode *ip, s64 xoff, int xlen, int flag);

extern int xtTailgate(
		      int tid,
		      struct inode *ip,
		      s64 xoff, int xlen, s64 xaddr, int flag);

extern int xtUpdate( int tid, struct inode *ip, struct xad *nxad);

extern int xtDelete(
		    int tid,
		    struct inode *ip, s64 xoff, int xlen, int flag);

extern int xtTruncate( int tid, struct inode *ip, s64 newsize, int type);

extern int xtRelocate(int tid,
		      struct inode *ip,
		      xad_t * oxad, s64 nxaddr, int xtype);

extern int xtAppend( int tid,	/* transaction id */
		    struct inode *ip,
		    int xflag, s64 xoff, int maxblocks, int * xlenp,	/* (in/out) */
		    s64 * xaddrp,	/* (in/out) */
		    int flag);

#ifdef  _JFS_DEBUG
int xtDisplayTree( struct inode *ip);

int xtDisplayPage( struct inode *ip, s64 bn, xtpage_t * p);
#endif				/* _JFS_DEBUG */
#endif				/* _JFS_UTILITY */

#endif				/* !_H_JFS_XTREE */
